#region Copyright (c) 2006-2008 Cellbi
/*
Cellbi Software Component Product
Copyright (c) 2006-2008 Cellbi
www.cellbi.com

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

	1.	Redistributions of source code must retain the above copyright notice,
			this list of conditions and the following disclaimer.

	2.	Redistributions in binary form must reproduce the above copyright notice,
			this list of conditions and the following disclaimer in the documentation
			and/or other materials provided with the distribution.

	3.	The names of the authors may not be used to endorse or promote products derived
			from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED AS IS AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CELLBI
OR ANY CONTRIBUTORS TO THIS SOFTWARE BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#endregion
//#pragma warning disable 3008
using System;
using System.IO;
using GetDocText.Native;
using GetDocText.Ole;
using System.Text;
using GetDocText.Doc;
using System.IO.Packaging;
using System.Collections.Generic;
using System.Xml;
using Word2007 = Microsoft.Office.Interop.Word;
using Word2000 = Office9.Word;
using Word2003 = Office11.Word;
using System.Linq;
using System.ComponentModel.Composition;
using CommonUtils;

namespace TextExtractors
{
    /// <summary>
    /// Implements loading of the text from the doc files.
    /// </summary>
    [Export(typeof(ITextImport))]
    [ExportMetadata("Description", "Word files")]
    [ExportMetadata("FileExtension", "*.doc; *.docx")]
    [ExportMetadata("ContextMenuText", "&Translate")]
    class DocReader : BaseTextExtractor, ITextImport
    {
        string _Path;

        // constructors ...
        /// <summary>
        /// Creates new instance of the TextLoader.
        /// </summary>
        /// <param name="path">The path of the file to load the text</param>
        public DocReader(string path)
        {
            _Path = path;
        }
        public DocReader()
        {

        }
        // private methods ...
        BinaryReader GetReader(OleStream stream)
        {
            if (stream == null)
                return null;
            byte[] streamData = stream.ReadToEnd();
            MemoryStream memoryStream = new MemoryStream(streamData);
            return new BinaryReader(memoryStream, Encoding.Unicode);
        }
        BinaryReader GetStreamReader(OleStorage storage, string streamName)
        {
            OleStream stream = storage.OpenStream(streamName);
            if (stream == null)
                return null;
            return GetReader(stream);
        }
        BinaryReader GetDocumentStreamReader(OleStorage storage)
        {
            return GetStreamReader(storage, "WordDocument");
        }
        BinaryReader GetTableStreamReader(OleStorage storage, string tableName)
        {
            return GetStreamReader(storage, tableName);
        }
        void GetDataFromFib(BinaryReader reader, out string tableName, out int pdcOffset, out uint pdcLength)
        {
            reader.BaseStream.Seek(10, SeekOrigin.Begin);
            ushort flags = reader.ReadUInt16();
            tableName = BitUtils.IsSet(flags, 9) ? "1Table" : "0Table";

            reader.BaseStream.Seek(418, SeekOrigin.Begin);
            pdcOffset = reader.ReadInt32();
            pdcLength = reader.ReadUInt32();
        }
        PieceDescriptorCollection GetPieceDescriptors(BinaryReader reader, int offset, uint length)
        {
            PieceDescriptorCollection result = new PieceDescriptorCollection(offset, length);
            result.Read(reader);
            return result;
        }

        static readonly Encoding encoder = new UnicodeEncoding();
        static readonly char[] filtered = new char[] { '\x0000', '\x0001',  '\x0002', '\x0003','\x0004', '\x0005','\x0006', '\x0007','\x0008', '\x0009','\x000B', '\x000C', '\x000E','\x000F',
        '\x0010', '\x0011',  '\x0012', '\x0013','\x0014', '\x0015','\x0016', '\x0017','\x0018', '\x0019', '\x001A','\x001B', '\x001C','\x001D', '\x001E','\x001F'};
        string ReadString(BinaryReader reader, uint length, bool isUnicode)
        {
            if (length == 0)
                return string.Empty;

            if (isUnicode)
                length = length / 2;
            StringBuilder result = new StringBuilder((int)length);

            try
            {

                if (isUnicode)
                {
                    byte[] raw = reader.ReadBytes((int)length);
                    char[] chars = encoder.GetChars(raw);
                    chars.Where(c => !filtered.Contains(c)).ToList().ForEach(ch => result.Append(ch));
                }
                else
                {
                    for (int i = 0; i < length; i++)
                    {
                        byte ch = reader.ReadByte();
                        if (ch > 31 || ch == 13 || ch == 10)
                            result.Append(CharFromByte(ch));
                    }
                }
                result.Append(Environment.NewLine);
                return result.ToString();
            }
            finally
            {
                result.Length = 0;
            }
        }
        Encoding _docEncoding;
        bool LoadText(OleStorage storage, out string text)
        {
            text = string.Empty;
            if (storage == null)
                return false;

            BinaryReader documentReader = GetDocumentStreamReader(storage);
            if (documentReader == null)
                return false;

            int pdcOffset;
            uint pdcLength;
            string tableName;
            GetDataFromFib(documentReader, out tableName, out pdcOffset, out pdcLength);

            BinaryReader tableReader = GetTableStreamReader(storage, tableName);
            if (tableReader == null)
                return false;

            PieceDescriptorCollection pieces = GetPieceDescriptors(tableReader, pdcOffset, pdcLength);
            if (pieces == null)
                return false;

            int count = pieces.Count;
            StringBuilder sb = new StringBuilder();
            var arg = new ProgressEventArgs();
            arg.Description = "OleStorage Import";
            for (int i = 0; i < count; i++)
            {
                uint pieceStart;
                uint pieceEnd;
                bool isUnicode = pieces.GetPieceFileBounds(i, out pieceStart, out pieceEnd);
                if (isUnicode)
                    _docEncoding = Encoding.UTF8;
                arg.Progress = ((float)i + 1) / count;
                if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                    ProgressChanged(this, arg);
                documentReader.BaseStream.Seek(pieceStart, SeekOrigin.Begin);
                sb.Append(ReadString(documentReader, pieceEnd - pieceStart, isUnicode));

            }
            text = sb.Replace("\r", Environment.NewLine).ToString();
            sb.Length = 0;
            return true;
        }

        // public methods ...
        /// <summary>
        /// Loads text from the file.
        /// </summary>
        /// <param name="text">The text of the file</param>
        public bool LoadText(out string text)
        {
            var arg = new ProgressEventArgs();
            arg.Description = "Using default LoadText to open .doc file";
            ProgressChanged(this, arg);
            text = string.Empty;
            if (NativeMethods.StgIsStorageFile(_Path) != 0)
                return false;

            OleStorage storage = OleStorage.CreateInstance(_Path);
            try
            {
                return LoadText(storage, out text);
            }
            finally
            {
                storage.Close();
            }
        }
        private static PackagePart GetDocumentXmlPart(Package package)
        {
            foreach (PackageRelationship docRel in package.GetRelationshipsByType(
                @"http://schemas.openxmlformats.org/officeDocument/2006/relationships/officeDocument"))
            {
                return package.GetPart(PackUriHelper.ResolvePartUri(
                    new Uri("/", UriKind.Relative), docRel.TargetUri));
            }
            return null;
        }
        static private string ReadNode(XmlNode node)
        {
            if (node == null || node.NodeType != XmlNodeType.Element)
                return string.Empty;

            StringBuilder sb = new StringBuilder();
            try
            {
                foreach (XmlNode child in node.ChildNodes)
                {
                    if (child.NodeType != XmlNodeType.Element) continue;

                    switch (child.LocalName)
                    {
                        case "t":                           // Text
                            sb.Append(child.InnerText.TrimEnd());

                            string space = ((XmlElement)child).GetAttribute("xml:space");
                            if (!string.IsNullOrEmpty(space) && space == "preserve")
                                sb.Append(' ');

                            break;

                        case "cr":                          // Carriage return
                        case "br":                          // Page break
                            sb.Append(Environment.NewLine);
                            break;

                        case "tab":                         // Tab
                            sb.Append("\t");
                            break;

                        case "p":                           // Paragraph
                            sb.Append(ReadNode(child));
                            sb.Append(Environment.NewLine);
                            //sb.Append(Environment.NewLine);
                            break;

                        default:
                            sb.Append(ReadNode(child));
                            break;
                    }
                }
                return sb.ToString();
            }
            finally
            {
                sb.Length = 0;
            }
        }
        private string LoadFromFile(string file)
        {
            var arg = new ProgressEventArgs();
            arg.Description = "Using default LoadFromFile to open .docx file";
            ProgressChanged(this, arg);
            using (Package package = Package.Open(file, FileMode.Open, FileAccess.Read))
            {
                PackagePart docXmlPart = GetDocumentXmlPart(package);
                if (docXmlPart != null)
                {
                    using (Stream documentXml = docXmlPart.GetStream())
                    {
                        XmlDocument xmlDoc = new XmlDocument();
                        xmlDoc.PreserveWhitespace = true;
                        xmlDoc.Load(documentXml);
                        documentXml.Close();
                        const string WordprocessingMlNamespace = @"http://schemas.openxmlformats.org/wordprocessingml/2006/main";

                        XmlNamespaceManager nsmgr = new XmlNamespaceManager(xmlDoc.NameTable);
                        nsmgr.AddNamespace("w", WordprocessingMlNamespace);
                        const string BodyXPath = "/w:document/w:body";
                        XmlNode node = xmlDoc.DocumentElement.SelectSingleNode(BodyXPath, nsmgr);

                        if (node == null)
                            return null;
                        else
                            return ReadNode(node);
                    }
                }
                else
                    return null;
            }
        }


        #region ITextImport Members
        //see http://omegacoder.com/?p=555
        public string GetAllTextFromFile(string fileName, out Encoding encoding)
        {
            _Path = fileName;
            string text = null;
            encoding = Encoding.UTF8;
            try
            {
                string extension = Path.GetExtension(fileName).ToLower();
                //try using word 2007
                text = ExtractUsingWord2007();
                if (!string.IsNullOrEmpty(text))
                    return text;
                if (".doc" == extension)
                {
                    //try using word 2003
                    text = ExtractUsingWord2003();
                    if (!string.IsNullOrEmpty(text))
                        return text;
                    //try using word 2000
                    text = ExtractUsingWord2000();
                    if (!string.IsNullOrEmpty(text))
                        return text;
                }


                //fallback if automation failed
                switch (extension)
                {
                    case ".doc":
                        LoadText(out text);
                        if (_docEncoding != null)
                            encoding = _docEncoding;
                        break;
                    case ".docx":
                        text = LoadFromFile(fileName);
                        break;
                }
            }
            catch (Exception ex)
            {

                var arg = new ProgressEventArgs();
                arg.Description = ex.Message;
                arg.Progress = 1.0f;
                System.Windows.Forms.MessageBox.Show(ex.StackTrace, ex.Message);
                ProgressChanged(this, arg);
            }
            return text;
        }
        /*
         * //#pragma warning disable 467
      private string ExtractUsingWord2007()
      {
          object objNull = System.Reflection.Missing.Value;
          Word2007.Application wordApp = null;
          StringBuilder sb = new StringBuilder();
          try
          {
              var arg = new ProgressEventArgs();
              arg.Description = "Opening Word Doc";
              arg.Progress = 0;
              if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                  ProgressChanged(this, arg);
              wordApp = new Word2007.Application();
              // Create obj filename to pass it as paremeter in open 
              object objFile = _Path;
              object objReadOnly = true;//Open Document 
              Word2007.Document doc = wordApp.Documents.Open(ref objFile, ref objNull, ref objReadOnly, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull);
              if (doc != null)
              {
                  arg.Description = "Word2007 Import";
                  int count = doc.Paragraphs.Count;
                  for (int i = 1; i <= count; i++)
                  {
                      sb.Append(doc.Paragraphs[i].Range.Text.Trim(new char[] { '\a', '\u0001' }));
                      arg.Progress = ((float)i) / count;
                      if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                          ProgressChanged(this, arg);
                  }
                  sb.Replace("\r", Environment.NewLine);
                  // close document and Quit Word Application
                  doc.Close(ref objNull, ref objNull, ref objNull);
                  return sb.ToString();
              }
          }
          catch (Exception)
          {

          }
          finally
          {
              if (wordApp != null)
              {
                  wordApp.Quit(ref objNull, ref objNull, ref objNull);
                  //System.Runtime.InteropServices.Marshal.ReleaseComObject(wordApp);
              }
              sb.Length = 0;
          }
          return null;
      }
      private string ExtractUsingWord2003()
      {
          object objNull = System.Reflection.Missing.Value;
          Word2003.Application wordApp = null;
          StringBuilder sb = new StringBuilder();
          try
          {
              var arg = new ProgressEventArgs();
              arg.Description = "Opening Word Doc";
              arg.Progress = 0;
              if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                  ProgressChanged(this, arg);
              wordApp = new Word2003.Application();
              // Create obj filename to pass it as paremeter in open 
              object objFile = _Path;
              object objReadOnly = true;//Open Document 
              Word2003.Document doc = wordApp.Documents.Open(ref objFile, ref objNull, ref objReadOnly, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull);
              if (doc != null)
              {
                  // To read each line consider each line as paragraph 
                  arg.Description = "Word2003 Import";
                  int count = doc.Paragraphs.Count;
                  //foreach (Word2003.Paragraph objParagraph in doc.Paragraphs)
                  for (int i = 1; i <= count; i++)
                  {
                      sb.Append(doc.Paragraphs[i].Range.Text.Trim(new char[] { '\a', '\u0001' }));
                      arg.Progress = ((float)i) / count;
                      if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                          ProgressChanged(this, arg);
                  }
                  sb.Replace("\r", Environment.NewLine);
                  // close document and Quit Word Application
                  doc.Close(ref objNull, ref objNull, ref objNull);
                  return sb.ToString();
              }
          }
          catch (Exception)
          {

          }
          finally
          {
              if (wordApp != null)
              {
                  wordApp.Quit(ref objNull, ref objNull, ref objNull);
                  //System.Runtime.InteropServices.Marshal.ReleaseComObject(wordApp);
              }
              sb.Length = 0;
          }
          return null;
      }

      private string ExtractUsingWord2000()
      {
          object objNull = System.Reflection.Missing.Value;
          Word2000.Application wordApp = null;
          StringBuilder sb = new StringBuilder();
          try
          {
              var arg = new ProgressEventArgs();
              arg.Description = "Opening Word Doc";
              arg.Progress = 0;
              if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                  ProgressChanged(this, arg);
              wordApp = new Word2000.Application();
              // Create obj filename to pass it as paremeter in open 
              object objFile = _Path;
              object objReadOnly = true;//Open Document 
              Word2000.Document doc = wordApp.Documents.Open(ref objFile, ref objNull, ref objReadOnly, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull);
              if (doc != null)
              {
                  // To read each line consider each line as paragraph 
                  arg.Description = "Word2000 Import";
                  int count = doc.Paragraphs.Count;
                  for (int i = 1; i <= count; i++)
                  {
                      sb.Append(doc.Paragraphs.Item(i).Range.Text.Trim(new char[] { '\a', '\u0001' }));
                      arg.Progress = ((float)i) / count;
                      if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                          ProgressChanged(this, arg);
                  }
                  sb.Replace("\r", Environment.NewLine);
                  // close document and Quit Word Application
                  doc.Close(ref objNull, ref objNull, ref objNull);
                  return sb.ToString();
              }
          }
          catch (Exception)
          {

          }
          finally
          {
              if (wordApp != null)
              {
                  wordApp.Quit(ref objNull, ref objNull, ref objNull);
                  //System.Runtime.InteropServices.Marshal.ReleaseComObject(wordApp);
              }
              sb.Length = 0;
          }
          return null;
      }
         *   //#pragma warning restore 467
        */


        //#pragma warning disable 467
        private string ExtractUsingWord2007()
        {
            object objNull = System.Reflection.Missing.Value;
            Word2007._Application wordApp = null;
            StringBuilder sb = new StringBuilder();
            var arg = new ProgressEventArgs();
            try
            {

                arg.Description = "Opening Word 2007 Doc";
                arg.Progress = 0;
                if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                    ProgressChanged(this, arg);
                wordApp = new Word2007.Application();
                // Create obj filename to pass it as paremeter in open 
                object objFile = _Path;
                object objReadOnly = true;//Open Document 
                //Word2007._Document doc = wordApp.Documents.Open(ref objFile, ref objNull, ref objReadOnly, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull);
                Word2007._Document doc = wordApp.Documents.Open(_Path, ReadOnly: true, Visible: false);
                if (doc != null)
                {
                    arg.Description = "Word2007 Import";
                    int count = doc.Paragraphs.Count;
                    for (int i = 1; i <= count; i++)
                    {
                        sb.Append(doc.Paragraphs[i].Range.Text.Trim(new char[] { '\a', '\u0001' }));
                        arg.Progress = ((float)i) / count * 100;
                        if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                            ProgressChanged(this, arg);
                    }
                    sb.Replace("\r", Environment.NewLine);
                    // close document and Quit Word Application
                    //wordApp.Documents.Close(SaveChanges: false, OriginalFormat: false, RouteDocument: false);
                    doc.Close(SaveChanges: false);
                    return sb.ToString();
                }
            }
            catch (Exception ex)
            {
                arg.Description = ex.Message;
                ProgressChanged(this, arg);
            }
            finally
            {
                if (wordApp != null)
                {
                    wordApp.Quit(SaveChanges: false, OriginalFormat: false, RouteDocument: false);
                    //System.Runtime.InteropServices.Marshal.ReleaseComObject(wordApp);
                }
                sb.Clear();
            }
            return null;
        }
        private string ExtractUsingWord2003()
        {
            object objNull = System.Reflection.Missing.Value;
            Word2003._Application wordApp = null;
            StringBuilder sb = new StringBuilder();
            var arg = new ProgressEventArgs();
            try
            {

                arg.Description = "Opening Word 2003 Doc";
                arg.Progress = 0;
                if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                    ProgressChanged(this, arg);
                wordApp = new Word2003.Application();
                // Create obj filename to pass it as paremeter in open 
                object objFile = _Path;
                object objReadOnly = true;//Open Document 
                //Word2003._Document doc = wordApp.Documents.Open(ref objFile, ref objNull, ref objReadOnly, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull);
                Word2003._Document doc = wordApp.Documents.Open(_Path, ReadOnly: true, Visible: false);
                if (doc != null)
                {
                    // To read each line consider each line as paragraph 
                    arg.Description = "Word2003 Import";
                    int count = doc.Paragraphs.Count;
                    //foreach (Word2003.Paragraph objParagraph in doc.Paragraphs)
                    for (int i = 1; i <= count; i++)
                    {
                        sb.Append(doc.Paragraphs[i].Range.Text.Trim(new char[] { '\a', '\u0001' }));
                        arg.Progress = ((float)i) / count * 100;
                        if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                            ProgressChanged(this, arg);
                    }
                    sb.Replace("\r", Environment.NewLine);
                    // close document and Quit Word Application
                    doc.Close(SaveChanges: false);
                    //wordApp.Documents.Close(SaveChanges: false, OriginalFormat: false, RouteDocument: false);
                    return sb.ToString();
                }
            }
            catch (Exception ex)
            {
                arg.Description = ex.Message;
                ProgressChanged(this, arg);
            }
            finally
            {
                if (wordApp != null)
                {
                    //wordApp.Quit(ref objNull, ref objNull, ref objNull);
                    wordApp.Quit(SaveChanges: false, OriginalFormat: false, RouteDocument: false);
                    //System.Runtime.InteropServices.Marshal.ReleaseComObject(wordApp);
                }
                sb.Clear();
            }
            return null;
        }

        private string ExtractUsingWord2000()
        {
            object objNull = System.Reflection.Missing.Value;
            Word2000._Application wordApp = null;
            StringBuilder sb = new StringBuilder();
            var arg = new ProgressEventArgs();
            try
            {

                arg.Description = "Opening Word 2000 Doc";
                arg.Progress = 0;
                if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                    ProgressChanged(this, arg);
                wordApp = new Word2000.Application();
                // Create obj filename to pass it as paremeter in open 
                object objFile = _Path;
                object objReadOnly = true;//Open Document 
                //Word2000._Document doc = wordApp.Documents.Open(ref objFile, ref objNull, ref objReadOnly, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull, ref objNull);
                Word2000._Document doc = wordApp.Documents.Open(_Path, ReadOnly: true, Visible: false);
                if (doc != null)
                {
                    // To read each line consider each line as paragraph 
                    arg.Description = "Word2000 Import";
                    int count = doc.Paragraphs.Count;
                    for (int i = 1; i <= count; i++)
                    {
                        sb.Append(doc.Paragraphs.Item(i).Range.Text.Trim(new char[] { '\a', '\u0001' }));
                        arg.Progress = ((float)i) / count * 100;
                        if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                            ProgressChanged(this, arg);
                    }
                    sb.Replace("\r", Environment.NewLine);
                    // close document and Quit Word Application
                    // wordApp.Documents.Close(SaveChanges: false, OriginalFormat: false, RouteDocument: false);
                    doc.Close(SaveChanges: false);
                    return sb.ToString();
                }
            }
            catch (Exception ex)
            {
                arg.Description = ex.Message;
                ProgressChanged(this, arg);
            }
            finally
            {
                if (wordApp != null)
                {
                    wordApp.Quit(SaveChanges: false, OriginalFormat: false, RouteDocument: false);
                    //System.Runtime.InteropServices.Marshal.ReleaseComObject(wordApp);
                }
                sb.Clear();
            }
            return null;
        }

        //#pragma warning restore 467


        public event EventHandler ProgressChanged;
        #endregion
    }
}
