﻿/*
 * Please leave this Copyright notice in your code if you use it
 * Written by Decebal Mihailescu [http://www.codeproject.com/script/articles/list_articles.asp?userid=634640]
 */
using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using iTextSharp.text.pdf;
using System.IO;
using System.Text;
using System.ComponentModel.Composition;
using CommonUtils;

namespace TextExtractors
{
    /// <summary>
    /// Parses a PDF file and extracts the text from it.
    /// </summary>
    [Export(typeof(ITextImport))]
    //[ExportMetadata("SpecialFormat", false)]
    [ExportMetadata("Description", "Pdf files")]
    [ExportMetadata("FileExtension", "*.pdf")]
    [ExportMetadata("ContextMenuText", "&Translate Pdf File")]
    class PDFParser : BaseTextExtractor, ITextImport
    {
        /// BT = Beginning of a text object operator 
        /// ET = End of a text object operator
        /// Td move to the start of next line
        ///  5 Ts = superscript
        /// -5 Ts = subscript

        #region Fields

        #region _numberOfCharsToKeep
        /// <summary>
        /// The number of characters to keep, when extracting text.
        /// </summary>
        private static int _numberOfCharsToKeep = 15;
        #endregion

        #endregion

        #region ExtractText
        /// <summary>
        /// Extracts a text from a PDF file.
        /// </summary>
        /// <param name="inFileName">the full path to the pdf file.</param>
        /// <param name="outFileName">the output file name.</param>
        /// <returns>the extracted text</returns>
        public bool ExtractText(string inFileName, string outFileName)
        {
            StreamWriter outFile = null;
            try
            {
                // Create a reader for the given PDF file
                PdfReader reader = new PdfReader(inFileName);
                //outFile = File.CreateText(outFileName);
                outFile = new StreamWriter(outFileName, false, System.Text.Encoding.UTF8);

                Console.Write("Processing: ");
                var arg = new ProgressEventArgs();
                arg.Description = "Opening Pdf file";
                arg.Progress = 0;
                if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                    ProgressChanged(this, arg);
                int totalLen = 68;
                float charUnit = ((float)totalLen) / (float)reader.NumberOfPages;
                int totalWritten = 0;
                float curUnit = 0;

                for (int page = 1; page <= reader.NumberOfPages; page++)
                {
                    outFile.Write(ExtractTextFromPDFBytes(reader.GetPageContent(page)) + " ");

                    // Write the progress.
                    if (charUnit >= 1.0f)
                    {
                        for (int i = 0; i < (int)charUnit; i++)
                        {
                            Console.Write("#");
                            totalWritten++;
                        }
                    }
                    else
                    {
                        curUnit += charUnit;
                        if (curUnit >= 1.0f)
                        {
                            for (int i = 0; i < (int)curUnit; i++)
                            {
                                Console.Write("#");
                                totalWritten++;
                            }
                            curUnit = 0;
                        }

                    }
                    if (_pageNumbers > 1)
                    {
                        arg.Description = "Pdf Import";
                        arg.Progress = ((float)page) / _pageNumbers;
                        if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                            ProgressChanged(this, arg);
                    }
                }

                if (totalWritten < totalLen)
                {
                    for (int i = 0; i < (totalLen - totalWritten); i++)
                    {
                        Console.Write("#");
                    }
                }
                return true;
            }
            catch
            {
                return false;
            }
            finally
            {
                if (outFile != null) outFile.Close();
            }
        }

        int _pageNumbers = 0;
        public bool ExtractText(string inFileName, out string result)
        {
            //StreamWriter outFile = null;
            result = null;
            StringBuilder sb = new StringBuilder();
            try
            {
                // Create a reader for the given PDF file
                PdfReader reader = new PdfReader(inFileName);

                //Console.Write("Processing: ");
                var arg = new ProgressEventArgs();
                arg.Description = "Opening Pdf file";
                arg.Progress = 0;
                if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                    ProgressChanged(this, arg);

                int totalLen = 68;
                float charUnit = ((float)totalLen) / (float)reader.NumberOfPages;
                int totalWritten = 0;
                float curUnit = 0;
                _pageNumbers = reader.NumberOfPages;
                for (int page = 1; page <= _pageNumbers; page++)
                {
                    sb.Append(ExtractTextFromPDFBytes(reader.GetPageContent(page)));
                    sb.Append(" ");

                    // Write the progress.
                    if (charUnit >= 1.0f)
                    {
                        for (int i = 0; i < (int)charUnit; i++)
                        {
                            Console.Write("#");
                            totalWritten++;
                        }
                    }
                    else
                    {
                        curUnit += charUnit;
                        if (curUnit >= 1.0f)
                        {
                            for (int i = 0; i < (int)curUnit; i++)
                            {
                                Console.Write("#");
                                totalWritten++;
                            }
                            curUnit = 0;
                        }

                    }
                    if (_pageNumbers > 1)
                    {
                        arg.Description = "Pdf Import";
                        arg.Progress = ((float)page) / _pageNumbers;
                        if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                            ProgressChanged(this, arg);
                    }
                }

                if (totalWritten < totalLen)
                {
                    for (int i = 0; i < (totalLen - totalWritten); i++)
                    {
                        Console.Write("#");
                    }
                }
                return true;
            }
            catch
            {
                sb.Length = 0;
                return false;

            }
            finally
            {
                result = sb.ToString();
                sb.Length = 0;
                //if (outFile != null) outFile.Close();
            }
        }
        #endregion

        #region ExtractTextFromPDFBytes
        static readonly Encoding encoder = new UnicodeEncoding();

        /// <summary>
        /// This method processes an uncompressed Adobe (text) object 
        /// and extracts text.
        /// </summary>
        /// <param name="input">uncompressed</param>
        /// <returns></returns>
        private string ExtractTextFromPDFBytes(byte[] input)
        {
            if (input == null || input.Length == 0) return "";
            StringBuilder resultBuilder = new StringBuilder(input.Length / 3);
            try
            {

                // Flag showing if we are we currently inside a text object
                bool inTextObject = false;

                // Flag showing if the next character is literal 
                // e.g. '\\' to get a '\' character or '\(' to get '('
                bool nextLiteral = false;

                // () Bracket nesting level. Text appears inside ()
                int bracketDepth = 0;
                
                // Keep previous chars to get extract numbers etc.:
                char[] previousCharacters = new char[_numberOfCharsToKeep];
                for (int j = 0; j < _numberOfCharsToKeep; j++) previousCharacters[j] = ' ';


                for (int i = 0; i < input.Length; i++)
                {
                    char c = (char)input[i];

                    if (inTextObject)
                    {
                        // Position the text
                        if (bracketDepth == 0)
                        {
                            if (CheckToken(new string[] { "TD", "Td" }, previousCharacters))
                            {
                                resultBuilder.Append(Environment.NewLine);// += "\n\r";
                            }
                            else
                            {
                                if (CheckToken(new string[] { "'", "T*", "\"" }, previousCharacters))
                                {
                                    resultBuilder.Append("\n");// += "\n";
                                }
                                else
                                {
                                    if (CheckToken(new string[] { "Tj" }, previousCharacters))
                                    {
                                        resultBuilder.Append(" ");// += " ";
                                    }
                                }
                            }
                        }

                        // End of a text object, also go to a new line.
                        if (bracketDepth == 0 &&
                            CheckToken(new string[] { "ET" }, previousCharacters))
                        {

                            inTextObject = false;
                            resultBuilder.Append(' ');// += " ";
                        }
                        else
                        {
                            // Start outputting text
                            if ((c == '(') && (bracketDepth == 0) && (!nextLiteral))
                            {
                                bracketDepth = 1;
                            }
                            else
                            {
                                // Stop outputting text
                                if ((c == ')') && (bracketDepth == 1) && (!nextLiteral))
                                {
                                    bracketDepth = 0;
                                }
                                else
                                {
                                    // Just a normal text character:
                                    if (bracketDepth == 1)
                                    {
                                        // Only print out next character no matter what. 
                                        // Do not interpret.
                                        if (c == '\\' && !nextLiteral)
                                        {
                                            nextLiteral = true;
                                        }
                                        else
                                        {

                                            if (nextLiteral && c >= '0' && c <= '9')
                                            {
                                                string octalCode = ((char)input[i]).ToString() + ((char)input[++i]).ToString() + ((char)input[++i]).ToString();
                                                int nn = Convert.ToInt32(octalCode, 8);

                                                resultBuilder.Append(CharFromByte((byte)nn));
                                                nextLiteral = false;
                                            }
                                            else if (((c >= ' ') && (c <= '~')) ||
                                                ((c >= 128) && (c < 255)))
                                            {
                                                //resultString += @"\" + Convert.ToString((short)c, 8);
                                                resultBuilder.Append(c);
                                            }
                                            else
                                            {
                                                switch (c)
                                                {
                                                    case '\n':
                                                        resultBuilder.Append(c);// += c.ToString();
                                                        break;
                                                    case '\r':
                                                        resultBuilder.Append(Environment.NewLine);// "\r\n";
                                                        break;
                                                    default:
                                                        resultBuilder.Append(c);
                                                        break;
                                                }

                                            }


                                            nextLiteral = false;
                                        }
                                    }
                                }
                            }
                        }
                    }

                    // Store the recent characters for 
                    // when we have to go back for a checking
                    for (int j = 0; j < _numberOfCharsToKeep - 1; j++)
                    {
                        previousCharacters[j] = previousCharacters[j + 1];
                    }
                    previousCharacters[_numberOfCharsToKeep - 1] = c;

                    // Start of a text object
                    if (!inTextObject && CheckToken(new string[] { "BT" }, previousCharacters))
                    {
                        inTextObject = true;
                    }
                    if (_pageNumbers < 2)
                    {
                        var arg = new ProgressEventArgs();
                        arg.Description = "Pdf Import";
                        arg.Progress = ((float)i + 1) / input.Length;
                        if (ProgressChanged != null && ProgressChanged.GetInvocationList().Length > 0)
                            ProgressChanged(this, arg);
                    }
                }
                return resultBuilder.ToString();
            }
            catch
            {
                return "";
            }
            finally
            {
            }
        }
        #endregion

        #region CheckToken
        /// <summary>
        /// Check if a certain 2 character token just came along (e.g. BT)
        /// </summary>
        /// <param name="search">the searched token</param>
        /// <param name="recent">the recent character array</param>
        /// <returns></returns>
        private bool CheckToken(string[] tokens, char[] recent)
        {
            foreach (string token in tokens)
            {
                if (token.Length > 1) // Otherwise the "If" fails 
                {
                    if ((recent[_numberOfCharsToKeep - 3] == token[0]) &&
                        (recent[_numberOfCharsToKeep - 2] == token[1]) &&
                        ((recent[_numberOfCharsToKeep - 1] == ' ') ||
                        (recent[_numberOfCharsToKeep - 1] == 0x0d) ||
                        (recent[_numberOfCharsToKeep - 1] == 0x0a)) &&
                        ((recent[_numberOfCharsToKeep - 4] == ' ') ||
                        (recent[_numberOfCharsToKeep - 4] == 0x0d) ||
                        (recent[_numberOfCharsToKeep - 4] == 0x0a))
                        )
                    {
                        return true;
                    }
                }
            }
            return false;
        }
        #endregion

        #region ITextImport Members

        public string GetAllTextFromFile(ref string fileName,out Encoding encoding)
        {
            string result;
            encoding = Encoding.UTF8;
            if (ExtractText(fileName, out result))
                return result;
            else
                return null;
        }

        //public string FileDescription
        //{
        //    get { return "Pdf files"; }
        //}

        //public string FileExtension
        //{
        //    get { return "*.pdf"; }
        //}
        public event EventHandler ProgressChanged;

        #endregion
    }
}
